local ffi = require('ffi')
local bor, band, lshift, rshift
do
  local _obj_0 = require('bit')
  bor, band, lshift, rshift = _obj_0.bor, _obj_0.band, _obj_0.lshift, _obj_0.rshift
end
local min, max, floor
do
  local _obj_0 = math
  min, max, floor = _obj_0.min, _obj_0.max, _obj_0.floor
end
local APU_cdef = [[	struct {
		struct { //  0x4000 .. 0x4003
			uint8_t ctrl;
			uint8_t sweep;
			uint8_t timer;
			uint8_t counter;
		} pulse1;
		struct { // 0x4004 .. 0x4007
			uint8_t ctrl;
			uint8_t sweep;
			uint8_t timer;
			uint8_t counter;
		} pulse2;
		struct { // 0x4008 .. 0x400b
			uint8_t timer;
			uint8_t counter;
			uint8_t unused;
			uint8_t linear;
		} triangle;
		struct { // 0x400c .. 0x400f
			uint8_t timer;
			uint8_t counter;
			uint8_t envelope;
			uint8_t lfsr;
		} noise;
		struct { // 0x4010 .. 0x4013
			uint8_t timer;
			uint8_t reader;
			uint8_t buffer;
			uint8_t output;
		} dmc;
		uint8_t dma; // 0x4014 - dma
		uint8_t status; // 0x4015
		uint8_t ctrl1;  // 0x4016 - forwarded to controllers
		uint8_t framecounter; // 0x4017 - only bits 6-7

		int _oddClock;
	}
]]
local LENGTHS = {
  [0] = 10,
  254,
  20,
  2,
  40,
  4,
  80,
  6,
  160,
  8,
  60,
  10,
  14,
  12,
  26,
  14,
  12,
  16,
  24,
  18,
  48,
  20,
  96,
  22,
  192,
  24,
  72,
  26,
  16,
  28,
  32,
  30
}
local PulseChannel
PulseChannel = function(regs)
  local counter, timer = 0, 0
  local halted = false
  return {
    setLength = function(self, l)
      counter = l
    end,
    getCounter = function(self)
      return counter
    end,
    setHalt = function(self, val)
      halted = val
    end,
    clock = function(self)
      if halted then
        return 
      end
      if counter > 0 then
        counter = counter - 1
        if counter == 0 then
          return nil
        end
      end
    end
  }
end
local TriangleChannel
TriangleChannel = function()
  return {
    clock = function(self) end
  }
end
local NoiseChannel
NoiseChannel = function()
  return {
    clock = function(self) end
  }
end
local DMC
DMC = function()
  return {
    clock = function(self) end
  }
end
return function(system)
  local MODE1 = 14915 * 2
  local MODE2 = 18642 * 2
  local cpu = nil
  local ctrl1, ctrl2 = nil, nil
  local ppu = system:getPPU()
  local counterMode = 0
  local irqEnabled = false
  local stepMode = nil
  local step = 0
  local frameCounter = 0
  local frameIRQ = false
  local resetTimer = false
  local pulse1, pulse2 = nil, nil
  local triangle = nil
  local noise = nil
  local SEQUENCER = {
    [0] = {
      function(self) end,
      function(self)
        return self:_stepLength()
      end,
      function(self) end,
      function(self)
        return self:_stepLength()
      end
    },
    [1] = {
      function(self) end,
      function(self) end,
      function(self) end,
      function(self) end,
      function(self) end
    }
  }
  local readReg = {
    [0x00] = function(self)
      return 0
    end,
    [0x01] = function(self)
      return 0
    end,
    [0x02] = function(self)
      return 0
    end,
    [0x03] = function(self)
      return 0
    end,
    [0x04] = function(self)
      return 0
    end,
    [0x05] = function(self)
      return 0
    end,
    [0x06] = function(self)
      return 0
    end,
    [0x07] = function(self)
      return 0
    end,
    [0x08] = function(self)
      return 0
    end,
    [0x09] = function(self)
      return 0
    end,
    [0x0a] = function(self)
      return 0
    end,
    [0x0b] = function(self)
      return 0
    end,
    [0x0c] = function(self)
      return 0
    end,
    [0x0d] = function(self)
      return 0
    end,
    [0x0e] = function(self)
      return 0
    end,
    [0x0f] = function(self)
      return 0
    end,
    [0x10] = function(self)
      return 0
    end,
    [0x11] = function(self)
      return 0
    end,
    [0x12] = function(self)
      return 0
    end,
    [0x13] = function(self)
      return 0
    end,
    [0x14] = function(self)
      return 0
    end,
    [0x15] = function(self)
      if frameIRQ then
        cpu:irq_high()
      end
      local status = self.status
      self:_setFrameIRQ(false)
      local p1c = min(1, pulse1:getCounter())
      local p2c = min(1, pulse2:getCounter())
      status = band(status, 0xfc) + p1c + lshift(p2c, 1)
      return status
    end,
    [0x16] = function(self)
      return ctrl1:read()
    end,
    [0x17] = function(self)
      return ctrl2:read()
    end,
    [0x18] = function(self)
      return 0
    end,
    [0x19] = function(self)
      return 0
    end,
    [0x1a] = function(self)
      return 0
    end,
    [0x1b] = function(self)
      return 0
    end,
    [0x1c] = function(self)
      return 0
    end,
    [0x1d] = function(self)
      return 0
    end,
    [0x1e] = function(self)
      return 0
    end,
    [0x1f] = function(self)
      return 0
    end
  }
  local writeReg = {
    [0x00] = function(self, val)
      return pulse1:setHalt(band(val, 0x20) > 0)
    end,
    [0x01] = function(self, val) end,
    [0x02] = function(self, val) end,
    [0x03] = function(self, val)
      local enabled = band(self.status, 0x01)
      return pulse1:setLength(LENGTHS[rshift(val, 3)] * enabled)
    end,
    [0x04] = function(self, val)
      return pulse2:setHalt(band(val, 0x20) > 0)
    end,
    [0x05] = function(self, val) end,
    [0x06] = function(self, val) end,
    [0x07] = function(self, val)
      local enabled = band(rshift(self.status, 1), 0x01)
      return pulse2:setLength(LENGTHS[rshift(val, 3)] * enabled)
    end,
    [0x08] = function(self, val) end,
    [0x09] = function(self, val) end,
    [0x0a] = function(self, val) end,
    [0x0b] = function(self, val) end,
    [0x0c] = function(self, val) end,
    [0x0d] = function(self, val) end,
    [0x0e] = function(self, val) end,
    [0x0f] = function(self, val) end,
    [0x10] = function(self, val) end,
    [0x11] = function(self, val) end,
    [0x12] = function(self, val) end,
    [0x13] = function(self, val) end,
    [0x14] = function(self, val)
      return ppu:dma(val)
    end,
    [0x15] = function(self, val)
      self.status = val
      if band(val, 0x01) == 0 then
        pulse1:setLength(0)
      end
      if band(val, 0x02) == 0 then
        return pulse2:setLength(0)
      end
    end,
    [0x16] = function(self, val)
      self.ctrl1 = val
      ctrl1:write(val)
      return ctrl2:write(val)
    end,
    [0x17] = function(self, val)
      self.framecounter = val
      counterMode = rshift(band(val, 0x80), 7)
      stepMode = counterMode == 0 and self._stepMode0 or self._stepMode1
      irqEnabled = band(val, 0x40) == 0
      if not (irqEnabled) then
        if frameIRQ then
          cpu:irq_high()
        end
        self:_setFrameIRQ(false)
      end
      resetTimer = true
    end,
    [0x18] = function(self, val) end,
    [0x19] = function(self, val) end,
    [0x1a] = function(self, val) end,
    [0x1b] = function(self, val) end,
    [0x1c] = function(self, val) end,
    [0x1d] = function(self, val) end,
    [0x1e] = function(self, val) end,
    [0x1f] = function(self, val) end
  }
  return ffi.metatype(ffi.typeof(APU_cdef), {
    __new = function(self)
      local apu = ffi.new(self)
      stepMode = apu._stepMode0
      pulse1 = PulseChannel(apu.pulse1)
      pulse2 = PulseChannel(apu.pulse2)
      triangle = TriangleChannel()
      noise = NoiseChannel()
      return apu
    end,
    __index = {
      setCPU = function(self, c)
        cpu = c
      end,
      setController = function(self, j1, j2)
        ctrl1, ctrl2 = j1, j2
      end,
      reset = function(self) end,
      step = function(self)
        stepMode(self)
        self._oddClock = band(self._oddClock + 1, 0x01)
      end,
      read = function(self, reg)
        return readReg[reg](self)
      end,
      write = function(self, reg, val)
        return writeReg[reg](self, val)
      end,
      _setFrameIRQ = function(self, val)
        frameIRQ = val
        if val then
          self.status = bor(self.status, 0x40)
        else
          self.status = band(self.status, 0xbf)
        end
      end,
      _updateCounter = function(self, length)
        frameCounter = frameCounter + 1
        if resetTimer and self._oddClock == 1 then
          if band(self.framecounter, 0x80) > 0 then
            self:_stepLength()
          end
          frameCounter = 0
          resetTimer = false
        end
        frameCounter = frameCounter % length
        local s = math.floor(frameCounter / 7457)
        if s ~= step and s > 0 then
          step = s
          return true
        end
      end,
      _stepMode0 = function(self)
        if self:_updateCounter(MODE1 + 3) then
          SEQUENCER[0][step](self)
        end
        if irqEnabled then
          if frameCounter > MODE1 - 1 then
            self:_setFrameIRQ(true)
          end
          if frameIRQ then
            return cpu:irq_low()
          end
        end
      end,
      _stepMode1 = function(self)
        if self:_updateCounter(MODE2 + 3) then
          return SEQUENCER[1][step](self)
        end
      end,
      _stepLength = function(self)
        pulse1:clock()
        return pulse2:clock()
      end
    }
  })()
end
