local ffi = require('ffi')
local bor, band, bxor, bnot, lshift, rshift
do
  local _obj_0 = require('bit')
  bor, band, bxor, bnot, lshift, rshift = _obj_0.bor, _obj_0.band, _obj_0.bxor, _obj_0.bnot, _obj_0.lshift, _obj_0.rshift
end
local min, max
do
  local _obj_0 = math
  min, max = _obj_0.min, _obj_0.max
end
local CPU_ct = ffi.typeof("\n	struct {\n		uint8_t A, X, Y, SP, P;\n		uint16_t PC;\n\n		uint8_t  _port[2]; // for 6510\n		int _ir_requested;   // 0 or irq address vector\n		int _ir_pending;   // 0 or irq address vector\n\n		int _irq_line;\n		int _nmi_line;\n\n		int16_t _cycles;\n		int8_t _delay_irq;\n		uint8_t _instruction; // next instruction\n		double _instCounter;\n		double _cycleCounter;\n	}\n")
local STACK = 0x100
local NMI = 0xfffa
local RESET = 0xfffc
local IRQ = 0xfffe
return function(mem, BCD)
  if BCD == nil then
    BCD = true
  end
  local state = CPU_ct()
  local OPCODES, INSTRUCTIONS, TIMING = nil, nil, nil
  local cpu = {
    state = state,
    step = function(self)
      self:_step()
      state._cycleCounter = state._cycleCounter + 1
    end,
    readPort = function(self, addr)
      return state._port[addr]
    end,
    writePort = function(self, addr, val)
      state._port[addr] = val
      return val
    end,
    push8 = function(self, val)
      mem:write8(STACK + state.SP, val)
      state.SP = state.SP - 1
    end,
    pop8 = function(self)
      state.SP = state.SP + 1
      return mem:read8(STACK + state.SP)
    end,
    push16 = function(self, val)
      self:push8(band(rshift(val, 8), 0xff))
      return self:push8(band(val, 0xff))
    end,
    pop16 = function(self, val)
      return self:pop8() + self:pop8() * 0x100
    end,
    setC = function(self, val)
      state.P = band(bor(state.P, val), 0xfe + val)
    end,
    setZ = function(self, val)
      state.P = band(bor(state.P, val), 0xfd + val)
    end,
    setI = function(self, val)
      state.P = band(bor(state.P, val), 0xfb + val)
    end,
    setD = function(self, val)
      state.P = band(bor(state.P, val), 0xf7 + val)
    end,
    setV = function(self, val)
      state.P = band(bor(state.P, val), 0xbf + val)
    end,
    setN = function(self, val)
      state.P = band(bor(state.P, val), 0x7f + val)
    end,
    updateV = function(self, op, p, m)
      self:setV(band(bxor(op, p), band(bxor(op, m), 0x80)) / 2)
      return op
    end,
    updateZ = function(self, op)
      return self:setZ(2 - min(op, 1) * 2)
    end,
    updateN = function(self, op)
      return self:setN(band(op, 0x80))
    end,
    updateZN = function(self, op)
      local p = state.P
      if op == 0 then
        p = band(bor(p, 0x02), 0x7f)
      else
        local b = band(op, 0x80)
        p = band(bor(band(p, 0xfd), b), 0x7f + b)
      end
      state.P = p
      return op
    end,
    testStatus = function(self, flag)
      return band(state.P, flag) > 0
    end,
    reset = function(self, addr, status)
      if status == nil then
        status = 0x20
      end
      do
        state.A, state.X, state.Y, state.SP, state.P = 0, 0, 0, 0xfd, status
        state.PC = addr or mem:read16(RESET)
        state._port[0], state._port[1] = 0x2f, 0x37
        state._ir_pending, state._ir_requested = 0, 0
        state._cycles, state._instCounter = 0, 0
        state._cycleCounter = 7
        self._step = self._fetch_cycle
        return state
      end
    end,
    irq_low = function(self)
      state._irq_line = 1
    end,
    irq_high = function(self)
      state._irq_line = 0
    end,
    nmi = function(self)
      state._ir_requested = NMI
    end,
    wait = function(self, cycles)
      state._cycles = state._cycles + cycles
    end,
    _pollInterrupts = function(self)
      do
        state._ir_pending = 0
        if state._irq_line > 0 and not self:testStatus(0x04) and state._ir_requested == 0 then
          state._ir_requested = IRQ
        end
        if state._ir_requested > 0 then
          state._ir_pending = state._ir_requested
          state._ir_requested = 0
          return state._instruction == 0
        end
      end
      return false
    end,
    _fetch_cycle = function(self)
      do
        if state._cycles > 0 then
          state._cycles = state._cycles - 1
          return 
        end
        if state._delay_irq > 0 then
          state._delay_irq = state._delay_irq - 1
        elseif state._ir_pending > 0 then
          self:_interrupt()
          state._ir_pending = 0
        end
      end
      self:_fetch()
      self._step = self._exec_cycle
    end,
    _exec_cycle = function(self)
      do
        if state._cycles > 2 then
          state._cycles = state._cycles - 1
          return 
        end
        state._cycles = 0
        if self:_pollInterrupts() then
          state._instruction = mem:read8(mem:read16(state._ir_pending))
        end
        INSTRUCTIONS[state._instruction](state)
        state._instCounter = state._instCounter + 1
        self._step = self._fetch_cycle
        return state
      end
    end,
    _fetch = function(self)
      do
        local b = mem:read8(state.PC)
        state._instruction = b
        state._cycles = state._cycles + TIMING[b]
        state.PC = state.PC + 1
        return state
      end
    end,
    _interrupt = function(self)
      do
        local addr
        addr, state._ir_pending = state._ir_pending, 0
        state._cycles = state._cycles + 7
        self:push16(state.PC)
        self:push8(band(state.P, 0xef))
        self:setI(0x04)
        state.PC = mem:read16(addr)
        return state
      end
    end
  }
  do
    OPCODES = require('CPU.opcodes')(cpu, state, mem, BCD)
    local inst
    do
      local _tbl_0 = { }
      for i = 0, 255 do
        _tbl_0[i] = OPCODES[i][4]
      end
      inst = _tbl_0
    end
    local timing = ffi.new('int8_t[255]')
    for i = 0, 255 do
      timing[i] = OPCODES[i][2]
    end
    INSTRUCTIONS, TIMING = inst, timing
  end
  return cpu
end
