local ffi = require('ffi')
local band, lshift, rshift
do
  local _obj_0 = require('bit')
  band, lshift, rshift = _obj_0.band, _obj_0.lshift, _obj_0.rshift
end
ffi.cdef([[	typedef struct __attribute__((__packed__)) {
		uint8_t header[4]; // NES\x1a
		uint8_t prg_rom_size_lsb;
		uint8_t chr_rom_size_lsb;
		uint8_t flags6;
		uint8_t flags7;
		uint8_t mapper_msb;
		uint8_t rom_size_msb;
		uint8_t prg_ram_size;
		uint8_t chr_ram_size;
		uint8_t timing;
		uint8_t system_type;
		uint8_t misc_roms;
		uint8_t expansion;
	} NESHeader_ct;
]])
local Cart
do
  local _class_0
  local _base_0 = {
    getSystem = function(self)
      return self._system
    end,
    getPRGROM = function(self)
      return self._prgROM
    end,
    getPRGRAM = function(self)
      return self._prgRAM
    end,
    getCHRROM = function(self)
      return self._chrROM
    end,
    getCHRRAM = function(self)
      return self._chrRAM
    end,
    getPRGROMSize = function(self)
      return self._prgROMSize
    end,
    getPRGRAMSize = function(self)
      return self._prgRAMSize
    end,
    getCHRROMSize = function(self)
      return self._chrROMSize
    end,
    getCHRRAMSize = function(self)
      return self._chrRAMSize
    end,
    getNTMirroring = function(self)
      return self._ntMirroring
    end,
    getMapper = function(self)
      return self._mapperId
    end,
    _loadROM = function(self, data)
      local header = ffi.cast('NESHeader_ct*', data)
      local version = rshift(band(header.flags7, 0x0c), 2)
      self._parsers[self:_detectFormat(header, 0)](self, header)
      print("ROM version", version)
      print("Mapper", self._mapperId)
      print("System", self._system)
      print("PRG-ROM", self._prgROMSize)
      print("PRG-RAM", self._prgRAMSize)
      print("CHR-ROM", self._chrROMSize)
      print("CHR-RAM", self._chrRAMSize)
      print("Name table mirroring", self._ntMirroring)
      if self._chrRAMSize > 0 then
        self._chrRAM = ffi.new('uint8_t[?]', self._chrRAMSize)
      end
      self._prgRAM = ffi.new('uint8_t[?]', self._prgRAMSize)
      return self:_loadData(data, 16 + (self._trainer and 0x200 or 0))
    end,
    _detectFormat = function(self, header, size)
      do
        local ver = band(header.flags7, 0x0c)
        local _exp_0 = ver
        if 0x08 == _exp_0 then
          return 'ines2'
        elseif 0x00 == _exp_0 then
          if header.timing + header.system_type + header.misc_roms + header.expansion == 0 then
            return 'ines'
          end
        end
      end
      return 'archaic'
    end,
    _loadData = function(self, data, offs)
      if self._prgROMSize > 0 then
        self._prgROM = ffi.new('uint8_t[?]', self._prgROMSize)
        ffi.copy(self._prgROM, data + offs, self._prgROMSize)
      end
      offs = offs + self._prgROMSize
      if self._chrROMSize > 0 then
        self._chrROM = ffi.new('uint8_t[?]', self._chrROMSize)
        return ffi.copy(self._chrROM, data + offs, self._chrROMSize)
      end
    end,
    _parsers = {
      archaic = function(self, header)
        print("WARNING: 'archaic' iNES format NYI. Using fallback")
        return Cart._parsers.ines(header)
      end,
      ines = function(self, header)
        do
          local _with_0 = header
          self._prgROMSize = (_with_0.prg_rom_size_lsb + band(_with_0.rom_size_msb, 0x0f) * 0x100) * 16384
          self._chrROMSize = (_with_0.chr_rom_size_lsb + band(_with_0.rom_size_msb, 0xf0) / 0x10) * 8192
          self._prgRAMSize = _with_0.prg_ram_size > 0 and _with_0.prg_ram_size * 8192 or 8192
          self._chrRAMSize = self._chrROMSize == 0 and 8192 or 0
          self._mapperId = band(_with_0.flags6, 0xf0) / 0x10 + band(_with_0.flags7, 0xf0)
          self._ntMirroring = band(_with_0.flags6, 0x09)
          self._system = (band(_with_0.system_type, 0x03) == 0x02 or band(_with_0.system_type, 0x01) == 0x01) and 'PAL' or 'NTSC'
          self._trainer = band(_with_0.flags6, 0x04) == 0x04
          return _with_0
        end
      end,
      ines2 = function(self, header)
        do
          local _with_0 = header
          self._prgROMSize = (_with_0.prg_rom_size_lsb + band(_with_0.rom_size_msb, 0x0f) * 0x100) * 16384
          self._chrROMSize = (_with_0.chr_rom_size_lsb + band(_with_0.rom_size_msb, 0xf0) / 0x10) * 8192
          self._prgRAMSize = _with_0.prg_ram_size > 0 and _with_0.prg_ram_size * 8192 or 8192
          self._chrRAMSize = _with_0.chr_ram_size * 8192
          self._mapperId = band(_with_0.flags6, 0xf0) / 0x10 + band(_with_0.flags7, 0xf0)
          self._ntMirroring = band(_with_0.flags6, 0x09)
          self._system = (band(_with_0.system_type, 0x03) == 0x02 or band(_with_0.system_type, 0x01) == 0x01) and 'PAL' or 'NTSC'
          self._trainer = band(_with_0.flags6, 0x04) == 0x04
          return _with_0
        end
      end
    }
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, rom)
      if not (rom:sub(1, 4) == 'NES\x1a') then
        error("Invalid ROM file")
      end
      self._prgROM, self._prgRAM, self._chrROM = nil, nil, nil
      self._prgROMSize, self._prgRAMSize, self._chrROMSize, self._chrRAMSize = 0, 0, 0, 0
      self._mapperId = 0
      self._trainer = false
      self._system = 'NTSC'
      self._ntMirroring = 0
      return self:_loadROM(ffi.cast('char*', rom))
    end,
    __base = _base_0,
    __name = "Cart"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Cart = _class_0
  return _class_0
end
