local ffi = require('ffi')
local band, bor, lshift, rshift
do
  local _obj_0 = require('bit')
  band, bor, lshift, rshift = _obj_0.band, _obj_0.bor, _obj_0.lshift, _obj_0.rshift
end
local min, max, abs
do
  local _obj_0 = math
  min, max, abs = _obj_0.min, _obj_0.max, _obj_0.abs
end
return function(nes)
  local mapper = require('Cart.Mapper')(nes)
  local cart = nes:getCart()
  local rom = cart:getPRGROM()
  local ram = cart:getPRGRAM()
  local chr = cart:getCHRROM() or cart:getCHRRAM()
  local prgSize = cart:getPRGROMSize()
  local prgBanks = ffi.new('int[2]', 0x0000, prgSize - 0x4000)
  local chrBanks = ffi.new('int[2]', 0x0000, 0x1000)
  local prgBankMode = 0
  local chrBankMode = 0
  local shiftRegister = 0
  local writeCount = 0
  local mirrorMode = 0
  local mirroring = {
    [0] = mapper.nametables.switch1L,
    [1] = mapper.nametables.switch1H,
    [2] = mapper.nametables.switch2V,
    [3] = mapper.nametables.switch2H
  }
  local switchPRG = {
    [0] = function(bank)
      local addr = band(bank, 0xfe) * 0x4000
      prgBanks[0] = addr
      prgBanks[1] = addr + 0x4000
    end,
    [1] = function(bank)
      return switchPRG[0](bank)
    end,
    [2] = function(bank)
      prgBanks[0] = 0x0000
      prgBanks[1] = bank * 0x4000
    end,
    [3] = function(bank)
      prgBanks[0] = bank * 0x4000
      prgBanks[1] = prgSize - 0x4000
    end
  }
  local switchCHR = {
    [0] = function(bank)
      bank = band(bank, 0xfe) * 0x1000
      chrBanks[0] = bank
      chrBanks[1] = bank + 0x1000
    end,
    [1] = function(bank)
      chrBanks[0] = bank * 0x1000
    end
  }
  local registers = {
    [0] = function(val)
      prgBankMode = band(rshift(val, 2), 0x03)
      chrBankMode = band(rshift(val, 4), 0x01)
      local mirror = band(val, 0x03)
      if mirror ~= mirrorMode then
        mirrorMode = mirror
        return mirroring[mirror]()
      end
    end,
    [1] = function(val)
      return switchCHR[chrBankMode](val)
    end,
    [2] = function(val)
      if chrBankMode == 0 then
        return 
      end
      chrBanks[1] = val * 0x1000
    end,
    [3] = function(val)
      local bank = band(val, 0x0f)
      return switchPRG[prgBankMode](band(val, 0x0f))
    end
  }
  local readPRG = {
    [0] = function(addr)
      return ram[addr]
    end,
    [1] = function(addr)
      return rom[prgBanks[0] + addr]
    end,
    [2] = function(addr)
      return rom[prgBanks[1] + addr]
    end
  }
  local writePRG = {
    [0] = function(addr, val)
      ram[addr - 0x6000] = val
    end,
    [1] = function(addr, val)
      if band(val, 0x80) > 0 then
        writeCount, shiftRegister = 0, 0
        return 
      end
      writeCount = writeCount + 1
      shiftRegister = bor(rshift(shiftRegister, 1), lshift(band(val, 0x01), 4))
      if writeCount < 5 then
        return 
      end
      local reg = rshift(addr - 0x8000, 13)
      registers[reg](shiftRegister)
      writeCount, shiftRegister = 0, 0
    end
  }
  mapper.write = function(self, addr, val)
    writePRG[rshift(addr, 15)](addr, val)
    return nil
  end
  mapper.read = function(self, addr)
    local bank = rshift(addr, 14) - 1
    return readPRG[bank](band(addr, 0x3fff))
  end
  mapper.readCHR = function(self, addr)
    local bank = rshift(addr, 12)
    return chr[chrBanks[bank] + band(addr, 0x0fff)]
  end
  return mapper
end
