local ffi = require('ffi')
local band, bor, lshift, rshift
do
  local _obj_0 = require('bit')
  band, bor, lshift, rshift = _obj_0.band, _obj_0.bor, _obj_0.lshift, _obj_0.rshift
end
local max, abs
do
  local _obj_0 = math
  max, abs = _obj_0.max, _obj_0.abs
end
return function(nes)
  local mapper = require('Cart.Mapper')(nes)
  local ppu = nes:getPPU()
  local cart = nes:getCart()
  local rom = cart:getPRGROM()
  local ram = cart:getPRGRAM()
  local chr = cart:getCHRROM() or cart:getCHRRAM()
  local prgSize = cart:getPRGROMSize()
  local chrSize = cart:getCHRROMSize() > 0 and cart:getCHRROMSize() or cart:getCHRRAMSize()
  local irqEnabled = false
  local irqCounter = 0
  local irqLatch = 0
  local reload = false
  local lastA12 = 0
  local lastA12fall = 0
  local prgBanks = ffi.new('int[4]', 0x8000, 0xa000, prgSize - 0x4000, prgSize - 0x2000)
  local chrBanks = ffi.new('int[8]', 0x0000, 0x0400, 0x0800, 0x0c00, 0x1000, 0x1400, 0x1800, 0x1c00)
  local prgMap = 0
  local bankSelect = 0
  local bankMode = 0
  local registers = {
    [0] = function(val)
      bankSelect = band(val, 0x07)
      prgMap = band(rshift(val, 6), 0x01)
      bankMode = rshift(band(val, 0x80), 5)
    end,
    [1] = function(val)
      if bankSelect > 5 then
        local b = band(band(val, 0x3f) * 0x2000, prgSize - 1)
        prgBanks[band(bankSelect, 0x01)] = b
        return 
      end
      if bankSelect > 1 then
        local b = band(val * 0x0400, chrSize - 1)
        chrBanks[bankSelect - (bankMode - 2)] = b
        return 
      end
      local b = band(band(val, 0xfe) * 0x0400, chrSize - 1)
      local bank = lshift(bankSelect, 1) + bankMode
      chrBanks[bank] = b
      chrBanks[bank + 1] = b + 0x0400
    end,
    [2] = function(val)
      if band(cart:getNTMirroring(), 0x04) == 0x04 then
        return 
      end
      if band(val, 0x01) == 0 then
        return mapper.nametables.switch2V()
      else
        return mapper.nametables.switch2H()
      end
    end,
    [3] = function(val) end,
    [4] = function(val)
      irqLatch = val
    end,
    [5] = function(val)
      irqCounter = 0
      reload = true
    end,
    [6] = function(val)
      irqEnabled = false
      return nes:getCPU():irq_high()
    end,
    [7] = function(val)
      irqEnabled = true
    end
  }
  local readPRG = {
    [0] = {
      [0] = function(addr)
        return ram[addr]
      end,
      [1] = function(addr)
        return rom[prgBanks[0] + addr]
      end,
      [2] = function(addr)
        return rom[prgBanks[1] + addr]
      end,
      [3] = function(addr)
        return rom[prgBanks[2] + addr]
      end,
      [4] = function(addr)
        return rom[prgBanks[3] + addr]
      end
    },
    [1] = {
      [0] = function(addr)
        return ram[addr]
      end,
      [1] = function(addr)
        return rom[prgBanks[2] + addr]
      end,
      [2] = function(addr)
        return rom[prgBanks[1] + addr]
      end,
      [3] = function(addr)
        return rom[prgBanks[0] + addr]
      end,
      [4] = function(addr)
        return rom[prgBanks[3] + addr]
      end
    }
  }
  local writePRG = {
    [0] = function(addr, val)
      ram[addr - 0x6000] = val
    end,
    [1] = function(addr, val)
      local reg = band(rshift(addr, 12), 0x06) + band(addr, 0x01)
      return registers[reg](val)
    end
  }
  local _clockCounter
  _clockCounter = function()
    irqCounter = max(0, irqCounter - 1)
    if reload then
      irqCounter, reload = irqLatch, false
    end
    if irqCounter == 0 then
      if irqEnabled then
        nes:getCPU():irq_low()
      end
      reload = true
    end
  end
  local _a12watch
  _a12watch = function(a12, cycle)
    local rise = false
    if a12 > lastA12 then
      rise = abs(cycle - lastA12fall) > 9
    elseif a12 < lastA12 then
      lastA12fall = cycle
    end
    lastA12 = a12
    if rise then
      return _clockCounter()
    end
  end
  mapper.write = function(self, addr, val)
    writePRG[rshift(addr, 15)](addr, val)
    return nil
  end
  mapper.read = function(self, addr)
    local bank = rshift(addr, 13) - 3
    return readPRG[prgMap][bank](band(addr, 0x1fff))
  end
  mapper.readCHR = function(self, addr)
    local bank = rshift(addr, 10)
    ppu:_updateA12(addr)
    return chr[chrBanks[bank] + band(addr, 0x03ff)]
  end
  local readVRAM = mapper.readVRAM
  mapper.readVRAM = function(self, addr)
    ppu:_updateA12(0)
    return readVRAM(self, addr)
  end
  nes:getPPU():setA12Watcher(_a12watch)
  return mapper
end
