return function(nes)
  local graphics
  graphics = love.graphics
  local screenImg = graphics.newImage(nes:getScreen())
  screenImg:setFilter('nearest')
  local screenCanvas = graphics.newCanvas(SCREEN_WIDTH, SCREEN_HEIGHT)
  local paletteImg, palette = require('PPU.Palette')(nes)
  local palShader = graphics.newShader([[		uniform Image pal;
		uniform vec3 emphasis;
		const float offset = 1.0 / 128.0;

		vec4 effect(vec4 col, Image tex, vec2 uv, vec2 fc) {
			float t = Texel(tex, uv).r * 4.0;
			vec4 c = Texel(pal, vec2(t + offset, .5));
			return vec4(col.rgb * emphasis * c.rgb, 1.0);
		}
	]])
  palShader:send('pal', paletteImg)
  palShader:send('emphasis', {
    1.0,
    1.0,
    1.0
  })
  local outputShader = graphics.newShader([[		uniform vec2 pixelSize;
		uniform vec2 aberration;

		const vec3 dark = vec3(.75, .75, .75);
		const vec3 bright = vec3(1.5, 1.5, 1.5);
		const vec3 RGB[3] = vec3[3](
			vec3(1.2, .9, .9),
			vec3(.9, 1.2, .9),
			vec3(.9, .9, 1.2)
		);

		vec4 effect(vec4 col, Image tex, vec2 uv, vec2 fc) {
			float oddline = mod(fc.y + .5, 2.0);
			vec3 modifier = RGB[int(mod(fc.x + fc.y, 3.0))];
			vec3 above = Texel(tex, vec2(uv.x, uv.y - pixelSize.y * .5)).rgb;
			vec3 c = modifier * vec3(
				Texel(tex, uv - pixelSize * oddline * aberration).r,
				Texel(tex, uv).g,
				Texel(tex, uv + pixelSize * oddline * aberration).b
			);

			return vec4(mix(c * dark, (c + above * .5), (1.0 - oddline)), 1.0);
		}
	]])
  local useShader = true
  local useFilter = true
  local hideOverscan = true
  do
    outputShader:send('pixelSize', {
      1 / screenCanvas:getWidth(),
      1 / screenCanvas:getHeight()
    })
    outputShader:send('aberration', {
      .5,
      .5
    })
  end
  return {
    getDimensions = function(self)
      return screenCanvas:getDimensions()
    end,
    capture = function(self)
      local pixels = { }
      nes:getScreen():mapPixel(function(x, y, r)
        pixels[y * 341 + x] = math.floor(r * 255 + .5)
        return r, r, r, r
      end)
      local t = love.image.newImageData(screenImg:getDimensions())
      t:mapPixel(function(x, y)
        local c = pixels[y * 341 + x]
        local r = palette[c * 3 + 1] / 255
        local g = palette[c * 3 + 2] / 255
        local b = palette[c * 3 + 3] / 255
        return r, g, b, 1
      end)
      return t
    end,
    toggleShader = function(self)
      useShader = not useShader
    end,
    toggleFilter = function(self)
      useFilter = not useFilter
      return screenCanvas:setFilter(useFilter and 'linear' or 'nearest')
    end,
    toggleOverscan = function(self)
      hideOverscan = not hideOverscan
    end,
    render = function(self, newScreenPixels)
      if newScreenPixels then
        screenImg:replacePixels(newScreenPixels)
      end
      do
        graphics.push('all')
        graphics.reset()
        self:_drawGame()
        local x, y, scale
        do
          local sx, sy = graphics.getDimensions()
          scale = math.min(sx / SCREEN_WIDTH, sy / SCREEN_HEIGHT)
          x, y = math.floor((sx - SCREEN_WIDTH * scale) * .5), math.floor((sy - SCREEN_HEIGHT * scale) * .5)
          x, y, scale = x, y, scale
        end
        graphics.setBlendMode('none')
        if useShader then
          graphics.setShader(outputShader)
        end
        graphics.translate(x, y)
        graphics.scale(scale)
        graphics.draw(screenCanvas)
        graphics.setShader()
        if hideOverscan then
          self:_drawOverscan()
        end
        graphics.pop()
        return x, y, scale
      end
    end,
    _drawGame = function(self)
      do
        graphics.setCanvas(screenCanvas)
        graphics.setShader(palShader)
        graphics.draw(screenImg)
        graphics.setShader()
        graphics.setCanvas()
        return graphics
      end
    end,
    _drawOverscan = function(self)
      do
        graphics.setColor(.2, .2, .2)
        graphics.rectangle('fill', 0, 0, SCREEN_WIDTH, 8)
        graphics.rectangle('fill', 0, SCREEN_HEIGHT - 8, SCREEN_WIDTH, 8)
        graphics.rectangle('fill', 0, 8, 8, SCREEN_HEIGHT - 16)
        graphics.rectangle('fill', SCREEN_WIDTH - 8, 8, 8, SCREEN_HEIGHT - 16)
        return graphics
      end
    end
  }
end
