local ffi = require('ffi')
local Mem = require('Mem.PPUMem')
local band, bor, rshift, lshift
do
  local _obj_0 = require('bit')
  band, bor, rshift, lshift = _obj_0.band, _obj_0.bor, _obj_0.rshift, _obj_0.lshift
end
local PPU_cdef = [[	struct {
		uint8_t ctrl;
		uint8_t mask;
		uint8_t status;
		uint8_t OAMaddr, OAM2addr;
		uint8_t data;

		uint16_t vramAddr, tempAddr;
		uint8_t fineX;

		uint8_t oam[256];
		uint8_t oam2[32];
		uint8_t sprites[64];

		uint16_t bgTable;
		uint16_t sprTable;

		int bgOn, spritesOn, renderingOn;
		int clipBG, clipSpr;

		uint16_t x, y;
		uint8_t dataBuffer;

		uint8_t bgTile;
		uint8_t bgAttr[3];
		uint16_t bgPat[3];
		uint8_t bgPatLo;

		int8_t sprSize;
		int8_t numSprites, numNextSprites;
		int8_t sprFlipH, sprFlipV;
		uint16_t sprAddr;
		int zeroHit;
		int spr0Active[2];
		int8_t oddFrame;

		uint8_t *frameBuffer;

		int cycleCount;
		int frameCount;
		int lastStatusRead;
		int nmiFired;
		int dmaAddr;
		int dmaCount;
		uint8_t dmaData;
		int writePhase;
	}
]]
local RASTERLEN = 341
local REGW = 1
local REGR = 2
local IRQ = 4
local LOG_FLAGS = 0
return function(nes)
  local STATES = require("PPU." .. tostring(nes:getCart():getSystem()) .. ".states")
  local pal = nes:isPAL()
  local LINES = pal and 313 or 262
  local RBIT = pal and 0x20 or 0x40
  local GBIT = pal and 0x40 or 0x20
  local mem = Mem()
  local screen = ffi.cast('uint8_t*', nes:getScreen():getFFIPointer())
  local fb = ffi.new('uint8_t[?]', SCREEN_WIDTH * SCREEN_HEIGHT)
  local nmiEnabled = false
  local addrIncr = 0x01
  local greyscale = false
  local eR, eG, eB = false, false, false
  local warmup = true
  local writeReg = {
    [0] = function(self, val)
      if warmup then
        return 
      end
      do
        local v = lshift(band(val, 0x03), 10)
        self.tempAddr = bor(band(self.tempAddr, 0x73ff), v)
      end
      self.ctrl = val
      local enableNMI = band(val, 0x80) > 0
      if enableNMI and not nmiEnabled and band(self.status, 0x80) > 0 then
        self.nmiFired = 0
      end
      nmiEnabled = enableNMI
      local largeSpr = rshift(band(val, 0x20), 5)
      self.bgTable = lshift(band(val, 0x10), 8)
      self.sprTable = rshift(band(val, 0x08), 3) * (1 - largeSpr)
      addrIncr = rshift(band(val, 0x04), 2) * 0x1f + 0x01
      self.sprSize = 8 + largeSpr * 8
    end,
    [1] = function(self, val)
      self.mask = val
      eB = band(val, 0x80) > 0
      eG = band(val, GBIT) > 0
      eR = band(val, RBIT) > 0
      greyscale = band(val, 0x01) > 0
      self.bgOn = band(rshift(val, 3), 0x01)
      self.spritesOn = band(rshift(val, 4), 0x01)
      self.renderingOn = bor(self.spritesOn, self.bgOn)
      self.clipSpr = (1 - self.spritesOn) * 512 + (0x08 - band(val, 0x04) * 2)
      self.clipBG = (1 - self.bgOn) * 512 + (0x08 - band(val, 0x02) * 4)
    end,
    [2] = function(self, val) end,
    [3] = function(self, val)
      self.OAMaddr = val
    end,
    [4] = function(self, val)
      if band(self.OAMaddr, 0x03) == 0x02 then
        val = band(val, 0xe3)
      end
      self.oam[self.OAMaddr] = val
      self.OAMaddr = self.OAMaddr + 1
    end,
    [5] = function(self, val)
      if self.writePhase == 0 then
        do
          local v = band(rshift(val, 3), 0x1f)
          self.tempAddr = bor(band(self.tempAddr, 0x7fe0), v)
          self.fineX = band(val, 0x07)
        end
      else
        do
          local v1 = lshift(band(val, 0x07), 12)
          local v2 = lshift(band(val, 0xf8), 2)
          self.tempAddr = bor(band(self.tempAddr, 0x0c1f), v1, v2)
        end
      end
      self.writePhase = band(self.writePhase + 1, 0x01)
    end,
    [6] = function(self, val)
      if self.writePhase == 0 then
        local v = lshift(band(val, 0x3f), 8)
        self.tempAddr = bor(band(self.tempAddr, 0xff), v)
      else
        self.tempAddr = bor(band(self.tempAddr, 0xff00), val)
        self.vramAddr = self.tempAddr
        self:_updateA12(self.vramAddr)
      end
      self.writePhase = band(self.writePhase + 1, 0x01)
    end,
    [7] = function(self, val)
      self.data = val
      mem.write(self.vramAddr, val)
      self.vramAddr = band(self.vramAddr + addrIncr, 0x3fff)
      return self:_updateA12(self.vramAddr)
    end
  }
  local readReg = {
    [0] = function(self)
      return 0
    end,
    [1] = function(self)
      return 0
    end,
    [2] = function(self)
      self.lastStatusRead = self.cycleCount
      local res = self.status
      self.status = band(res, 0x7f)
      self.writePhase = 0
      return res
    end,
    [3] = function(self)
      return 0
    end,
    [4] = function(self)
      local val = self.oam[self.OAMaddr]
      return val
    end,
    [5] = function(self)
      return 0
    end,
    [6] = function(self)
      return 0
    end,
    [7] = function(self)
      local addr
      addr, self.data = self.vramAddr, self.dataBuffer
      self.vramAddr = band(addr + addrIncr, 0x3fff)
      self:_updateA12(self.vramAddr)
      self.dataBuffer = mem.read(addr)
      if addr > 0x3eff then
        self.data = self.dataBuffer
        self.dataBuffer = mem.readPAL(0x2f00 + band(addr, 0x1f))
      end
      return self.data
    end
  }
  local state = STATES[0]
  local a12default
  a12default = function() end
  local a12watch = a12default
  return ffi.metatype(ffi.typeof(PPU_cdef), {
    __new = function(self)
      local ppu = ffi.new(self)
      return ppu
    end,
    __index = {
      log = function(self, grp, fmt, ...)
        if band(LOG_FLAGS, grp) == 0 then
          return 
        end
        return io.write(("%5d %04x %5d (%3d, %3d) | %s\n"):format(self.frameCount, nes:getCPU().state.PC, self.cycleCount, self.x, self.y, fmt:format(...)))
      end,
      step = function(self)
        if self.dmaCount > 0 then
          self:_serviceDMA()
        end
        return self:_tick(3)
      end,
      reset = function(self)
        ffi.fill(self, ffi.sizeof(self), 0)
        self.frameBuffer = fb
        nmiEnabled = false
        addrIncr = 1
        state = STATES[0]
        warmup = true
      end,
      getMem = function(self)
        return mem
      end,
      getState = function(self)
        return self.x, self.y
      end,
      present = function(self)
        ffi.copy(screen, self.frameBuffer, SCREEN_WIDTH * SCREEN_HEIGHT)
        return nes:nextFrame()
      end,
      nmi = function(self)
        if nmiEnabled and self.dmaCount == 0 then
          return nes:getCPU():nmi()
        end
      end,
      read = function(self, reg)
        return readReg[reg](self)
      end,
      write = function(self, reg, val)
        return writeReg[reg](self, val)
      end,
      dma = function(self, addr)
        self.dmaAddr = lshift(addr, 8)
        local latency = bor(self.oddFrame, self.renderingOn)
        self.dmaCount = 513 + latency
        return nes:getCPU():wait(513 + latency)
      end,
      setA12Watcher = function(self, watcher)
        a12watch = watcher or a12default
      end,
      _updateA12 = function(self, addr)
        return a12watch(band(rshift(addr, 12), 0x01), self.cycleCount)
      end,
      _tick = function(self, n)
        if n == 0 then
          return 
        end
        state[self.x](self, mem)
        self.cycleCount = self.cycleCount + 1
        self.x = self.x + 1
        if self.x < RASTERLEN then
          return self:_tick(n - 1)
        end
        self.x, self.y = 0, (self.y + 1) % LINES
        state = STATES[self.y]
        if self.y > 0 then
          return self:_tick(n - 1)
        end
        warmup = false
        self.frameCount = self.frameCount + 1
        self.oddFrame = band(self.oddFrame + 1, 0x01)
        self.cycleCount = 0
        return self:_tick(n - 1)
      end,
      _serviceDMA = function(self)
        self.dmaCount = self.dmaCount - 1
        if self.dmaCount < 512 then
          if band(self.dmaCount, 0x01) == 1 then
            self.dmaData = nes:getBus():read8(self.dmaAddr + 255 - rshift(self.dmaCount, 1))
          else
            return writeReg[0x04](self, self.dmaData)
          end
        end
      end
    }
  })()
end
