local ffi = require('ffi')
local graphics, filesystem, timer
do
  local _obj_0 = love
  graphics, filesystem, timer = _obj_0.graphics, _obj_0.filesystem, _obj_0.timer
end
local NES = require('NES')
local Display = require('Display')
local font = graphics.newFont('assets/fonts/RetroGaming.ttf', 12, 'none')
font:setFilter('nearest')
local setFrameCallback
setFrameCallback = function(nes, display)
  local capturedFrames = 0
  return nes:setFrameCallback(function()
    if __MOONSCRIPT and love.keyboard.isDown('lshift') and love.keyboard.isDown('lctrl') then
      local s = display:capture()
      s:encode('png', ("%04d.png"):format(capturedFrames))
      capturedFrames = capturedFrames + 1
    end
  end)
end
local loadROM
loadROM = function(file, isdata)
  local nfs = require('nativefs')
  local cart = isdata and file or assert(nfs.read(file))
  local nes = NES(cart)
  nes:reset()
  local display = Display(nes)
  setFrameCallback(nes, display)
  return nes, display
end
local nes, display
do
  if not __MOONSCRIPT then
    if arg[2] then
      nes, display = loadROM(arg[2])
    else
      nes, display = loadROM(filesystem.read('assets/defaultgame/2048.nes'), true)
    end
  else
    nes, display = loadROM(require('testrom'))
  end
end
local flushAll
flushAll = function()
  jit.flush()
  jit.off()
  collectgarbage()
  collectgarbage()
  return jit.on()
end
love.filedropped = function(file)
  if not (file:open('r')) then
    return 
  end
  local data = file:read()
  file:close()
  if data:sub(1, 4) == 'PK\x03\x04' then
    local fdata = filesystem.newFileData(data, 'rom.zip')
    filesystem.mount(fdata, 'rom')
    local fname = assert(filesystem.getDirectoryItems('/rom')[1])
    data = assert(filesystem.read("/rom/" .. tostring(fname)))
    filesystem.unmount('rom.zip')
  end
  nes, display = nil, nil
  flushAll()
  nes, display = loadROM(data, true)
end
local paused = false
love.gamepadpressed = function(joystick, button)
  return nes:gamepadpressed(joystick, button)
end
love.gamepadreleased = function(joystick, button)
  return nes:gamepadreleased(joystick, button)
end
love.gamepadaxis = function(joystick, axis, value)
  return nes:gamepadaxis(joystick, axis, value)
end
love.keypressed = function(key, scan)
  local _exp_0 = key
  if 'f1' == _exp_0 then
    return display:toggleShader()
  elseif 'f2' == _exp_0 then
    return display:toggleFilter()
  elseif 'f3' == _exp_0 then
    return display:toggleOverscan()
  elseif 'f5' == _exp_0 then
    return love.event.quit('restart')
  elseif 'f10' == _exp_0 then
    flushAll()
    return nes:getCPU():reset()
  elseif 'f12' == _exp_0 then
    _G.DEBUG = not _G.DEBUG
  elseif 'p' == _exp_0 then
    paused = not paused
  elseif 'escape' == _exp_0 then
    return love.event.quit()
  else
    return nes:keypressed(scan)
  end
end
love.keyreleased = function(key, scan)
  return nes:keyreleased(scan)
end
local accum = 0
local tickDuration = 0
local dbgTextX, dbgTextY, dbgTextScale = 0, 0, 2
local debugDraw
local draw
draw = function(dt)
  do
    graphics.clear(0, 0, 0, 0)
    graphics.line(0, 0, 0, 0)
    graphics.setFont(font)
  end
  accum = accum + dt
  if paused or accum <= 1 / 60 then
    dbgTextX, dbgTextY, dbgTextScale = display:render()
  else
    collectgarbage('stop')
    tickDuration = nes:tick()
    collectgarbage('restart')
    dbgTextX, dbgTextY, dbgTextScale = display:render(nes:getScreen())
    accum = math.max(0, accum - (1 / 60)) % (1 / 60)
  end
  if DEBUG then
    debugDraw(dt, tickDuration)
  end
  return graphics.present()
end
love.run = function()
  timer.step()
  return function()
    local event
    event = love.event
    event.pump()
    for name, a, b, c, d, e, f in event.poll() do
      if name == 'quit' then
        return a or 0
      end
      love.handlers[name](a, b, c, d, e, f)
    end
    local dt = timer.step()
    if graphics.isActive() then
      draw(math.min(1 / 20, dt))
    end
    if dt < 1 / 59 then
      timer.sleep(.001)
    end
    return nil
  end
end
do
  local COLOR_LABEL = {
    .5,
    .6,
    .7,
    1
  }
  local COLOR_INFO = {
    1,
    1,
    1,
    1
  }
  local COLOR_WARNING = {
    1,
    .2,
    0,
    1
  }
  local DBG_BARS
  do
    local _tbl_0 = { }
    for i = 0, 10 do
      _tbl_0[i] = ("|"):rep(i)
    end
    DBG_BARS = _tbl_0
  end
  local LOAD_COLORS = {
    [0] = {
      0,
      1,
      0,
      1
    },
    [100] = {
      1,
      .2,
      0,
      1
    }
  }
  local frame = 0
  local getLoadColor
  getLoadColor = function(pct)
    pct = math.floor(math.max(0, math.min(100, pct)) + .5)
    do
      local c = LOAD_COLORS[pct]
      if c then
        return c
      end
    end
    local v = pct * .01
    local c = {
      v,
      math.min(1, 1.2 - v * v),
      0,
      1
    }
    LOAD_COLORS[pct] = c
    return c
  end
  debugDraw = function(dt, tickDuration)
    do
      local text
      do
        local loadPct = 100 * (tickDuration / (1 / 60))
        local numPctBars = math.min(10, math.ceil(loadPct / 10))
        local fps = timer.getFPS()
        local garbage = collectgarbage('count')
        text = {
          COLOR_LABEL,
          "FPS:",
          COLOR_INFO,
          ("%3d "):format(fps),
          COLOR_LABEL,
          "MEM:",
          COLOR_INFO,
          ("%5.1fM "):format(garbage / 1024),
          COLOR_LABEL,
          "T:",
          getLoadColor(loadPct),
          ("%3.0f"):format(tickDuration * 1000),
          COLOR_LABEL,
          ("ms %s "):format(paused and "(paused)" or ""),
          getLoadColor(loadPct),
          DBG_BARS[numPctBars],
          {
            .4,
            .4,
            .4,
            1
          },
          DBG_BARS[10 - numPctBars]
        }
      end
      local tx, ty = dbgTextX + 8 * dbgTextScale, math.max(0, dbgTextY + (8 * (dbgTextScale - 2)))
      for i = 0, 1 do
        graphics.setColor(i, i, i)
        graphics.printf(text, tx + i, ty + i, graphics.getWidth() * 2, 'left')
      end
      frame = frame + 1
      return graphics
    end
  end
end
